{-
Ideas: can I implement application conditions in the rewrite rules?
For example: 
evalDiv (Con a :/: Con 0) = fail "Divide by zero"
evalDiv (Con a :/: Con b) = Just $ Con (a/b)
evalDiv _ = Nothing

input is transformation step, + selections (paths). For `normal' rewrite rules
I need a single selection, but for substitution I need two. 

In the UI: select a transformation rule, and then select as many expressions
as required by the transformation rule.

-}

module ApplyStrategy where

import Maybe 
import Data.Generics hiding ((:*:),Prefix)

import StrategyLib
import TermRep

--------------------------------------------------------------------------------
-- Rewriting
--------------------------------------------------------------------------------

class Apply symbol zipper_domain where
  apply :: symbol -> zipper_domain -> Maybe zipper_domain

class Children domain where
  children :: domain -> [domain]

instance Apply Char String   where
  apply c s  = Just s


--------------------------------------------------------------------------------
-- Locations and contexts
--------------------------------------------------------------------------------

data Loc domain context = At { it  :: domain
                             , ctx :: context} deriving Show

--------------------------------------------------------------------------------
-- Strategies
--------------------------------------------------------------------------------

{- The data type Strategy basic represents strategies
for rewriting using basic rules.
-}

data FBStrategy domain  =  RepeatExhausted (FBStrategy domain)
                        |  Topdown (FBStrategy domain)
                        |  Bottomup (FBStrategy domain)
                        |  EveryWhere (FBStrategy domain)
                        |  Or (FBStrategy domain) (FBStrategy domain)
                        |  Seq (FBStrategy domain) (FBStrategy domain)
                        |  Basic Rule

data Rule = Eval (TP Maybe) -- forall x. x -> Maybe x

--------------------------------------------------------------------------------
-- Applying strategies
--------------------------------------------------------------------------------

{-
applyStrategy :: Match domain => FBStrategy domain -> domain -> Maybe domain 
applyStrategy strategy expr = case strategy of
  RepeatExhausted strategy' -> case applyStrategy strategy' expr of
                                 Just expr' -> applyStrategy strategy expr'
                                 Nothing    -> Just expr
  Topdown strategy'         -> case applyStrategy strategy' expr of
                                 Just expr' -> Just expr'
                                 Nothing    -> mapChildrenM (applyStrategy strategy) expr
  EveryWhere strategy'      -> case applyStrategy strategy' expr of
                                 Just expr' -> Just (mapChildren (fromJust . applyStrategy strategy) expr')
                                 Nothing -> Just (mapChildren (fromJust . applyStrategy strategy) expr) 
                                 -- doesn't maintain the invariant that Just denotes rewritten.
  Basic rule                -> applyRule rule expr                                 
-}
--testSimplify = applyStrategy simplifyVExpr (Con 3 :+: ((Con 5 :-: Con 7) :*: Var "x"))

  
--------------------------------------------------------------------------------
-- Utility functions for parsing strategies
--------------------------------------------------------------------------------

{- Function applicable takes a strategy and a state, and tries to apply
the strategy to the state. All possible results are returned.
-}

applicable :: Data state => FBStrategy state -> state -> [Maybe state]
applicable strategy  state = case strategy of
  Basic (Eval f)  ->  case applyS f state of
                        Just state' -> [Just state']
                        Nothing       -> [Nothing]
  Or s1 s2    ->  applicable s1 state ++ applicable s2 state
{-  Seq s1 s2   ->  ( concatMap (applicable s2) 
                  . map fromJust 
                  . filter isJust 
                  ) (applicable s1 state)
  Topdown s   ->  let toprewrites = filter isJust (applicable s state)
                  in if    null toprewrites
                     then  concatMap (applicable strategy) (map fst (children state))
                     else  toprewrites
-}
{- firstRules determines the rules that can appear
as first `symbols' in the strategy.
-}

firstRules                      :: FBStrategy basic -> [Rule]
firstRules (Or s1 s2)           =  firstRules s1 ++ firstRules s2
firstRules (Basic rule)         =  [rule]
firstRules (RepeatExhausted s)  =  firstRules s


