module ParserCombinatorsWithStateBFLib where

--Remove the rsult type? Or make it always unit?

import Maybe
import List hiding (insert, delete)
import Debug.Trace

import Data.Generics

import ApplyStrategy

--trace _ v = v
infixl 3 <|>
infixl 4 <*> 
-- DERIVED COMBINATORS =================================================================
--infixl 5 <..>
infixl 4 <$>, <$, <*, *>, <**>, <??>
infixl 2 `opt`
-- INTERFACE ===========================================================================
type Accum    = Int
type Location = [String]
type Messages = String
data Steps c  = Step Int (Steps c)
              | Done Messages c
              deriving Show

resultOf (Step _ rest)      = resultOf rest
resultOf (Done msgs res)    = ( msgs, res)

data Tag a = Applicable a 
           | NotApplicable a 
           | Empty a

-- == PARSER TYPES =================================================================================================
 --                                  inp     m           l           st     acc       res
newtype Parser s st1 st2  a = P (forall b c .
                                 ([s] ->  Messages -> Location -> st2 -> Accum  -> b         -> Steps c) 
                              ->   [s] ->  Messages -> Location -> st1 -> Accum  -> (a -> b) -> Steps c
                                )
                                

unP (P p) = p
--  
pUpdateLoc :: (Location -> Location) -> Parser s st st Location
pUpdateLoc f = P (\k inp m l st acc res  -> let newloc   = f l  in k inp m  newloc st acc (res newloc)  )

pUpdate :: (st1 -> st2) -> Parser s st1 st2 st2
pUpdate    f = P (\k inp m l st acc res  -> let newstate = f st in k inp m l newstate acc (res newstate))

-- == AUXILIARY FUNCTIONS FOR CONSTRUCTING THE RESULT===================================
insert  a loc m  = m++"Inserted:" ++ show a++" at loc" ++ show loc ++ "\n"
delete  b loc m  = m++"Deleted :" ++ show b++" at loc" ++ show loc ++ "\n"
-- == MERGING STEPS ====================================================================
best x@(Done _ _)         _                  =  x
best _                y@(Done _ _)           =  y
best x@(Step l ls)    y@(Step r rs) 
  =  --trace ("Best of: " ++ show l ++ " and " ++ show r ++ "\n") $
     if      l <= r  then Step l (ls  `best`  y )
                     else Step r (x   `best`  rs)
-- ================= PARSERS ===========================================================
(<|>) :: Parser s st1 st2 a         -> Parser s st1 st2 a  -> Parser s st1 st2 a 
(P p) <|> (P q)    = P (\ k inp m l st acc res->  p k inp m l st acc res `best` q k inp m l st acc res)

(<*>) :: Parser s st1 st2 (b -> a)  -> Parser s st2 st3 b  -> Parser s st1 st3 a 
(P p) <*> (P q)    = P (\k inp m l st acc res -> p (q k) inp m l st acc (res.) )

pSucceed :: a -> Parser s st st a
pSucceed v         = P (\k inp m l st acc res -> k inp m l st acc (res v))

-- It is probably nicer to split pSym in a select and apply part
-- Alternatively: call this one pSym0or1, and also introduce a pSym2 for substitute.
pSym  :: (Apply s st,Show st,Symbol s) =>  s -> Parser s st st s 
pSym a =
   P (\k inp m l st acc res
 -> let insertres   = let newcost = acc + penalty a
                      in Step newcost (unP (pSym a) k (a:inp) (insert a l m) l st newcost res)
    in case inp of
       (s:ss) -> if isMove s 
                 then case apply s st of
                        Just st' -> Step acc (k ss m l st' acc (res s))
                        Nothing -> let newcost = acc + penalty s in Step newcost (unP (pSym a) k ss (delete s l m) l st newcost res)
                 else
                 let delresult   = let newcost = acc + penalty s
                                   in Step newcost (unP (pSym a) k ss (delete s l m) l st newcost res)                      
                     bothresults = insertres `best`  delresult                 
                 in case apply s st
                    of Just st'  -> if  a == s
                                    then  Step acc (k ss m l st' acc (res s))
                                          `best` 
                                          delresult
                                    else  bothresults 
                       Nothing -> bothresults
       []      ->  insertres `best` if null m 
                                    then Step acc (Done "prefix" (error "prefix")) 
                                    else Step acc (Done (m++"end-of-input") (error (show st)))   
     )

{- accept a symbol if it fits
pSym  :: (Apply s st,Show st,Symbol s) =>  s -> Parser s st st s 
pSym a =
   P (\k inp m l st acc res
 -> let insertres   = let newcost = acc + penalty a
                      in Step newcost (unP (pSym a) k (a:inp) (insert a l m) l st newcost res)
    in case inp of
       (s:ss) -> if isMove s 
                 then case  apply s st of
                        Just st' -> Step acc (k ss m l st' acc (res s))
                        Nothing -> let newcost = acc + penalty s in Step newcost (unP (pSym a) k ss (delete s l m) l st newcost res)
                 else
                 let delresult   = let newcost = acc + penalty s
                                   in Step newcost (unP (pSym a) k ss (delete s l m) l st newcost res)                      
                     bothresults = insertres `best`  delresult                 
                 in case apply s st
                    of Just st'  -> if a == s
                                    then  Step acc (k ss m l st' acc (res s))
                                    else  bothresults 
                       Nothing -> bothresults
       []      ->  insertres `best` if null m 
                                    then Step acc (Done "prefix" (error "prefix")) 
                                    else Step acc (Done (m++"end-of-input") (error (show st)))   
     )
-}
     
pFail :: Parser s st1 st2 a 
pFail      = P (\ k inp m l st acc res -> Step (acc + 10000) (unP pFail k inp m l st (acc + 10000) res))


-- PARSING Streams of Characters AND PRINTING RESULTS =========================================================
-- SYMBOL CLASS
class (Eq a, Show a) => Symbol a where
 penalty :: a -> Int
 isMove :: a -> Bool
 symQed :: a
 symPrefix :: a

instance Symbol Char where
 penalty s  = if s == '\EOT'  then 1000 else 5

-- parse :: (Match TExpr st,Show a, Show st, Show loc) => Parser TExpr (st,loc) a  -> [TExpr] -> (st,loc) -> IO ()
parse  p inp mystate = do let (P final)       = (p <* pSym symQed)
                          let steps           = resultOf (final (\inp m l st acc res -> if null inp 
                                                                                        then Done m res 
                                                                                        else Step (sum (map penalty inp)) (Done (m++"nonempty input:"++show inp) res)
                                                                ) inp "" [] mystate 0 id)
                          putStr (show steps)
                          let (msgs, val)     = steps
                          putStr (if null msgs then "" else "Errors:\n" ++ msgs)
                          putStr ("Result:\n"++show val ++['\n'])


{-
instance Symbol TExpr where
  penalty s = if s==QED then 1000 else 5
-}
{-
parse :: (Show a, Show st) => Parser Char (st,loc) a  -> [Char] -> (st,loc) -> IO ()
parse  p inp mystate = do let (P final)       = (p <* pSym '\EOT')
                          let (msgs, steps)   = final (\(st,res) -> ("", Done (st,res)))  (Default (inp++['\EOT']) mystate, id)
--                          putStr (show steps)
                          putStr (if null msgs then "" else "Errors:\n" ++ msgs)
                          putStr ("Result:\n"++show (resultOf steps)++['\n'])
-}
-- =======================================================================================
-- ===== DERIVED COMBINATORS =================== =========================================
-- =======================================================================================
opt :: Parser s st1 st1 a -> a -> Parser s st1 st1 a
(<$>)  :: (b -> a) -> Parser s st1 st2 b -> Parser s st1 st2 a
(<$)   :: (b -> a) -> Parser s st st x -> Parser s st st (b -> a)
(<*)   :: Parser s st1 st2 a  -> Parser s st2 st3 b        -> Parser s st1 st3 a
(*>)   :: Parser s st1 st2 b  -> Parser s st2 st3 a        -> Parser s st1 st3 a
(<**>) :: Parser s st1 st2 b  -> Parser s st2 st3 (b -> a) -> Parser s st1 st3 a
(<??>) :: Parser s st1 st2 a  -> Parser s st2 st2 (a -> a) -> Parser s st1 st2 a
p `opt` v       = p <|> pSucceed v   
f <$> p         = pSucceed f     <*> p
f <$  p         = const f        <$> p
p <*  q         = (\ x _ -> x)   <$> p <*> q
p *>  q         = (\ _ x -> x)   <$> p <*> q
p <**> q        = (\ x f -> f x) <$> p <*> q
p <??> q        =                    p <**> (q `opt` id)

pPacked :: Parser s st1 st2 x -> Parser s st3 st4 y -> Parser s st2 st3 a -> Parser s st1 st4 a
pPacked l r x   =   l *>  x <*   r


-- =======================================================================================
-- ===== Strategy parsers ===============================================================
-- =======================================================================================

pNot :: (Symbol t,Data a, Show a, Apply t (Loc a b),Children a) => FBStrategy a -> Parser t (Loc a b) (Loc a b) t
pNot strategy =    P (\k inp m l st acc res -> if  trace (show (notApplicableAnyWhere strategy (it st))) $ notApplicableAnyWhere strategy (it st)
                                               then trace (show inp) $ k inp m l st acc (res symQed) 
                                               else Step 1000 (Done "pNot" (error "pNot"))
                     )

notApplicableTop strategy expr = null $ filter isJust $ applicable strategy expr
notApplicableAnyWhere strategy expr = notApplicableTop strategy expr && 
                                      let children_expr = children expr
                                      in if null children_expr
                                         then True
                                         else and (map (notApplicableAnyWhere strategy) children_expr)

pRepeatExhausted :: (Symbol t,Data a,Show a,Apply t (Loc a b),Children a)  
                    => Parser t (Loc a b) (Loc a b) c
                    -> FBStrategy a
                    -> Parser t (Loc a b) (Loc a b) t
pRepeatExhausted p strategy  =  -- pUpdateLoc ("Entering pRepeatExhausted":) *>
                             pRepeatExhausted' p strategy
                                -- <* pUpdateLoc tail
  where pRepeatExhausted' p strategy  =  pNot strategy 
                                     <|> (p *> pRepeatExhausted p strategy)

-- =======================================================================================
-- ===== Iterating parsers ===============================================================
-- =======================================================================================
-- pFoldr      :: (a -> b -> b, b)               -> Parser s st a -> Parser s st b
-- pFoldr1     :: (a -> b -> b, b)               -> Parser s st a -> Parser s st b
-- pFoldrSep   :: (a -> b -> b, b) -> Parser s st x -> Parser s st a -> Parser s st b
-- pFoldr1Sep  :: (a -> b -> b, b) -> Parser s st x -> Parser s st a -> Parser s st b
-- pFoldr      alg@(op,e)     p =  (op <$> p <*> pFoldr alg p) `opt` e
-- pFoldr1     alg@(op,e)     p = op <$> p <*> pFoldr     alg p
-- pFoldrSep   alg@(op,e) sep p = ((op <$> p <*> pFoldr alg (sep *> p)) `opt` e         )
-- pFoldr1Sep  alg@(op,e) sep p =  pfm where pfm = op <$> p <*> pFoldr alg (sep *> p)
-- 
-- list_alg = ((:), [])
-- pList       :: Parser s st a               -> Parser s st [a]
-- pList1      :: Parser s st a               -> Parser s st [a]
-- pListSep    :: Parser s st b -> Parser s st a -> Parser s st [a]
-- pList1Sep   :: Parser s st b -> Parser s st a -> Parser s st [a]
-- pList        p = pFoldr     list_alg   p
-- pList1       p = pFoldr1    list_alg   p
-- pListSep   s p = pFoldrSep  list_alg s p
-- pList1Sep  s p = pFoldr1Sep list_alg s p
-- pChainr op x   =  r where r = x <??> (flip <$> op <*> r)
-- pChainl op x   =  r where  r      = (f <$> x <*> pList (flip <$> op <*> x) )
--                            f x [] = x
--                            f x (func:rest) = f (func x) rest
-- pAnyOf l       = foldr (<|>) pFail ( map pSym l)

-- (l <..> h)  k inp@(b:bs) | l <= b && b <= h = addstep 0.addresult b.k $ bs
--                         | otherwise = (insert l) (k inp)  `best` (delete b) ((l <..> h) k bs)
-- (l <..> h)  k inp@[]                 = (insert l) (k inp)
-- DEMONSTRATION PARSERS ==================================================================

lift x = [x]
a = pSym 'a'
b = pSym 'b'
c = pSym 'c'
d = pSym 'd'
-- S -> abcd | da
abcd = (\a b c d -> [a,b,c,d]) <$> a <*> b <*> c <*> d
da   = (\d a     -> [d,a    ]) <$>d <*> a
s = abcd <|> da

ps = parse s "a" "no-state"
-- bc = (\ x y -> x:[y]) <$> b <*> c
-- d1 = parse (concat <$> pList (lift <$> a <|> bc))  "abcdab" "my-own-state"
-- d2 = parse pXML "<a><b></c></a>" "no-state"

-- =======================================================================================
-- ===== MONADIC INTERFACE ===============================================================
-- =======================================================================================
-- monads werken niet omdat de state nu ook kan veranderen

(P p)  `andthen` q =  P (\ k inp  m  l  st  acc  res ->  
                        p (\  inp' m' l' st' acc' pv  -> unP (q pv) k inp' m' l' st' acc' res) inp m l st acc id
                        )


-- Testing Monadic interface
{-
pXML = do t <- pOpenTag
          Tag t <$> pList pXML <* pCloseTag t

data XML = Tag Char [XML] deriving Show

pTagged = pPacked (pSym '<') (pSym '>')
pOpenTag = pTagged pAnyChar
pCloseTag t = pTagged (pSym '/' *> pSym t)

pAnyChar = P (\k (st, res) -> case splitState st of 
                              Just (x, newstate) -> if x >= 'a' && x <= 'z' then
                                                       trace ("Probably Accepting " ++ show x ++ ['\n']) 
                                                       (addstep 0 (k (newstate, res x)))
                                                    else ("no letter in tag", alwaysFail)
                              Nothing            -> error "no character in tag"
             )

-}